/* libdf3 - open.c */

/* cf fdopen(3) */
#define _POSIX_C_SOURCE 200112L

#include "df3_priv.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <unistd.h>


static df3Error checkHeader(df3Handle *df3, off_t const dsz, int const fd) {
  endianHeader u = {{0}};
  if (DF3_HEADERSIZE != read(fd, &u, DF3_HEADERSIZE)) {
    df3->syserr = errno;
    return (df3->ec = ecSysErr);
  }

  df3Axes const a = {.x = ntohs(u.v[0]),
                     .y = ntohs(u.v[1]),
                     .z = ntohs(u.v[2])};
  off_t const csz = (off_t)AXMUL(a);

  if (0 == csz)
    return (df3->ec = ecNotDf3File);
  else if (csz == dsz)
    df3->bpv = 1;
  else if (2 * csz == dsz)
    df3->bpv = 2;
  else if (4 * csz == dsz)
    df3->bpv = 4;
  else
    return (df3->ec = ecNotDf3File);

  df3->axes = a;
  return (df3->ec = ecOk);
}


static inline bool emptyHandle(df3Handle *df3) {
  return !(df3->chcks  ||  df3->ec  ||  df3->dp  ||  df3->readonly
          ||  df3->ismapped  ||  df3->bpv
          ||  df3->axes.x  ||  df3->axes.y  ||  df3->axes.z);
}


// TODO rewrite with label_fail, cf create.
df3Error df3_open_file(df3Handle *df3, bool const ro) {
  /* expect only fname */
  if (!df3  ||  !emptyHandle(df3))
    return ecBadHandle;
  else if (!df3_check_fname(df3->fname))
    return (df3->ec = ecUnusableFile);

  off_t fsz = 0;
  if (0 != (df3->syserr = df3ExistFileCheck(df3->fname, &fsz)))
    return (df3->ec = ecSysErr);
  else if (DF3_HEADERSIZE >= fsz)
    return (df3->ec = ecNotDf3File);

  int const fd = open(df3->fname, ro ? O_RDONLY : O_RDWR);
  if (-1 == fd) {
    df3->syserr = errno;
    return (df3->ec = ecSysErr);
  }
  else if (ecOk != checkHeader(df3, fsz - DF3_HEADERSIZE, fd)) {
    (void)close(fd);
    return df3->ec;
  }
  else if (-1 == lseek(fd, 0, SEEK_SET)) {
    df3->syserr = errno;
    (void)close(fd);
    return (df3->ec = ecSysErr);
  }

  switch (df3->bpv) {
    case 1:  /* fall through if large file */
      if ((off_t)DF3_MMAPLIMIT >= fsz) {
        mmFile *p = malloc(sizeof(mmFile));
        if (!p) {
          df3->syserr = errno;
          (void)close(fd);
          return (df3->ec = ecSysErr);
        }
        int const prot = ro ? PROT_READ : (PROT_READ|PROT_WRITE);
        p->fd = fd;
        p->n = (size_t)fsz;
        p->dp = mmap(NULL, p->n, prot, MAP_SHARED, p->fd, 0);
        if (MAP_FAILED == p->dp) {
          df3->syserr = errno;
          (void)close(fd);
          free(p);
          return (df3->ec = ecSysErr);
        }
        df3->dp = p;
        df3->ismapped = true;
        break;
      }
    case 2:
    case 4: {
        FILE *fp = fdopen(fd, ro ? "rb" : "rb+");
        if (!fp) {
          df3->syserr = errno;
          (void)close(fd);
          return (df3->ec = ecSysErr);
        }
        df3->dp = fp;
        df3->ismapped = false;
      }
      break;
  }

  df3->readonly = ro;
  df3->chcks = df3CalcChecksum(df3);

  return (df3->ec = ecOk);
}

/* ----------------------------------------------------------------- *
 *  this file is part of 'libdf3', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.    *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.              *
 *  all rights reserved.                                             *
 * ----------------------------------------------------------------- */
